import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/foundation.dart';
import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_profile/ocg_profile.dart';

class ProfileFirebaseRepository extends ProfileRepository {
  final FirebaseFirestore _firestore;
  final TokenRepository tokenRepository;
  final IDHelper idHelper;

  static const String profiles = 'profiles';
  static const String followings = 'followings';

  CollectionReference get profileCollection => _firestore.collection(profiles);

  ProfileFirebaseRepository(
    this.idHelper,
    this.tokenRepository,
  ) : this._firestore = FirebaseFirestore.instance;

  @override
  Future<AppResponse<String>> create(Profile object) async {
    try {
      assert(object.userId != null);
      String id = idHelper.generateId();
      Profile profile = object.copyProfile(
        id: id,
      );
      var oldResponse = await getByUser(object.userId);
      if (oldResponse.isSuccess) {
        return AppResponse.named(error: "${oldResponse.error}");
      }
      var json = _toJson(profile);
      await profileCollection.doc(id).set(json);
      await tokenRepository.saveToken(AppToken.named(userId: id));
      return AppResponse.named(data: id);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> delete(String id) async {
    throw UnimplementedError();
  }

  Map<String, dynamic> _toJson(Profile profile) {
    var json = profile.toJson();
    json["lowerName"] = profile.name?.toLowerCase();
    return json;
  }

  @override
  Future<AppResponse<List<Profile>>> getList(
      {int page = 1, int count = 10}) async {
    try {
      var snapshot = await profileCollection.limit(count * page).get();
      var documents =
          snapshot.docs.map((e) => e.data()).skip((page - 1) * count).toList();
      var result = (await Future.wait(documents.map((e) async {
        return await _parseJson(e);
      })))
          .toList()
          .where((element) => element != null)
          .toList();
      return AppResponse.named(data: result);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  Future<Profile> _parseJson(
    profileJson,
  ) async {
    try {
      var newJson = profileJson ?? {};
      return Profile.fromJson(newJson);
    } catch (e) {
      return null;
    }
  }

  @override
  Future<AppResponse<Profile>> read(
    String id,
  ) async {
    var document = await profileCollection.doc(id).get();
    var data = document.data();
    return AppResponse.named(
        data: await _parseJson(
      data,
    ));
  }

  String validateUsername(String username) {
    if (InputHelper.isBlank(username)) {
      return "username blank";
    }
    if (username.contains(" ")) {
      return "Username can not contain spaces";
    }
    return null;
  }

  @override
  Future<AppResponse<bool>> update(Profile object) async {
    try {
      if (object.username != null) {
        String validateError = validateUsername(object.username);
        if (validateError != null) {
          return AppResponse.named(error: "$validateError");
        }
        var response = await getByUsername(object.username);
        if (response.isSuccess && response.data?.id != object.id) {
          return AppResponse.named(error: "Username exists");
        }
      }
      await profileCollection.doc(object.id).update(_toJson(object));
      return AppResponse.named(data: true);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  Future<AppResponse<Profile>> getByUsername(String username) async {
    try {
      var documents = await profileCollection
          .where("username", isEqualTo: username)
          .limit(1)
          .get();
      if (documents.docs.length == 1) {
        var profile = await _parseJson(documents.docs[0].data());
        if (profile != null)
          return AppResponse.named(data: profile);
        else {
          return AppResponse.named(error: "Not exists");
        }
      }
      return AppResponse.named(error: "Not exists");
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<Profile>> getByUser(String userId) async {
    try {
      var snapshot = await profileCollection
          .where("userId", isEqualTo: userId)
          .limit(1)
          .get();
      if (snapshot.docs.length > 0) {
        return AppResponse.named(
            data: await _parseJson(snapshot.docs[0].data()));
      } else {
        return AppResponse.named(error: "Not exists");
      }
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<Profile>> getCurrent() async {
    try {
      AppResponse<AppToken> tokenResponse = await tokenRepository.getToken();
      if (tokenResponse.isSuccess) {
        return await read(
          tokenResponse.data?.userId,
        );
      } else {
        return AppResponse.named(error: tokenResponse.error);
      }
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<String>> getCurrentId() async {
    AppResponse<AppToken> tokenResponse = await tokenRepository.getToken();
    if (tokenResponse.isSuccess) {
      return AppResponse.named(data: tokenResponse?.data?.userId);
    }
    return AppResponse.named(error: tokenResponse.error);
  }

  @override
  Future<AppResponse<bool>> follow(String to, {bool follow = true}) async {
    try {
      var token = await getCurrentId();
      if (!token.isSuccess) {
        return AppResponse.named(error: "Not logged in");
      }
      String from = token.data;
      ProfileFollowing following = ProfileFollowing.named(
          from: from, to: to, recordedAt: DateTime.now());
      String id = getId(from, to);
      if (follow)
        await _firestore.collection(followings).doc(id).set(following.toJson());
      else {
        await _firestore.collection(followings).doc(id).delete();
      }
      return AppResponse.named(data: true);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  String getId(String a, String b) {
    return "$a$b";
  }

  @override
  Stream<ProfileFollowing> onFollowerChanged() async* {
    var response = await getCurrentId();
    if (response.isSuccess) {
      String id = response.data;
      var query = _firestore
          .collection(followings)
          .where("to", isEqualTo: id)
          .orderBy("recordedAt", descending: true)
          .limit(1)
          .snapshots();
      await for (var snapshot in query) {
        var docs = snapshot.docs;
        if (docs.length == 1) {
          var data = docs[0].data();
          var profileFollowing = ProfileFollowing.fromJson(data);
          yield profileFollowing;
        }
      }
    }
  }

  @override
  Stream<ProfileFollowing> onFollowingChanged() async* {
    var response = await getCurrentId();
    if (response.isSuccess) {
      String id = response.data;
      var query = _firestore
          .collection(followings)
          .where("from", isEqualTo: id)
          .orderBy("recordedAt", descending: true)
          .limit(1)
          .snapshots();
      await for (var snapshot in query) {
        var docs = snapshot.docs;
        if (docs.length == 1) {
          var data = docs[0].data();
          var profileFollowing = ProfileFollowing.fromJson(data);
          yield profileFollowing;
        }
      }
    }
  }

  @override
  Future<AppResponse<List<ProfileFollowing>>> getFollowersFollowing(
      String id) async {
    try {
      List<ProfileFollowing> result = [];
      await Future.wait([
        Future(() async {
          var snapshot = await _firestore
              .collection(followings)
              .where("from", isEqualTo: id)
              .get();
          var docs = snapshot.docs
              .map((e) => ProfileFollowing.fromJson(e.data()))
              .toList();
          result.addAll(List<ProfileFollowing>.from(docs));
        }),
        Future(() async {
          var snapshot = await _firestore
              .collection(followings)
              .where("to", isEqualTo: id)
              .get();
          var docs = snapshot.docs
              .map((e) => ProfileFollowing.fromJson(e.data()))
              .toList();
          result.addAll(List<ProfileFollowing>.from(docs));
        })
      ]);
      return AppResponse.named(data: result);
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<Profile>>> getFollowers(String id) async {
    try {
      var snapshot = await _firestore
          .collection(followings)
          .where("to", isEqualTo: id)
          .get();
      var docs = snapshot.docs
          .map((e) => ProfileFollowing.fromJson(e.data()))
          .toList();
      var followersList = List<ProfileFollowing>.from(docs);
      var ids = List<String>.from(followersList.map((e) => e.from).toList());
      var profiles = await Future.wait<Profile>(ids.map((e) async {
        var response = await read(e);
        if (response.isSuccess) {
          return response.data;
        }
        return null;
      }).toList());
      profiles = profiles.where((element) => element != null).toList();
      return AppResponse.named(data: profiles);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<Profile>>> getFollowing(String id) async {
    try {
      var snapshot = await _firestore
          .collection(followings)
          .where("from", isEqualTo: id)
          .get();
      var docs = snapshot.docs
          .map((e) => ProfileFollowing.fromJson(e.data()))
          .toList();
      var followersList = List<ProfileFollowing>.from(docs);
      var ids = List<String>.from(followersList.map((e) => e.to).toList());
      var profiles = await Future.wait<Profile>(ids.map((e) async {
        var response = await read(e);
        if (response.isSuccess) {
          return response.data;
        }
        return null;
      }).toList());
      profiles = profiles.where((element) => element != null).toList();
      return AppResponse.named(data: profiles);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<Profile>>> getFriends(String userId) async {
    try {
      var response = await getFollowersFollowing(userId);
      if (response.isSuccess) {
        List<String> ids = List<String>.from(response.data.map((e) {
          if (e.from == userId) {
            return e.to;
          } else if (e.to == userId) {
            return e.from;
          }
        }));
        var result =
            await Future.wait<Profile>(Set<String>.from(ids).map((e) async {
          var response = await read(e);
          return response.data;
        }).toList());
        result = result.where((element) => element != null).toList();
        return AppResponse.named(data: result);
      }
      return AppResponse.named(error: response.error);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }
}
