import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:ocg_app/ocg_app.dart';
import 'package:http/http.dart';
import 'package:ocg_profile/src/models/profile.dart';
import 'package:ocg_profile/src/repository/base.dart';

class ProfileNetworkRepository extends ProfileRepository {
  final BaseClient client;
  final TokenRepository tokenRepository;

  ProfileNetworkRepository(this.tokenRepository, this.client);

  @override
  Future<AppResponse<List<Profile>>> getList(
      {int count = 10, int page = 1}) async {
    try {
      String url = "/contacts";
      var response = await client.get(
        url,
      );
      var json = jsonDecode(response.body);
      List<String> contacts = List<String>.from(json["contacts"] as List);
      Set<Profile> users = Set<Profile>();
      await Future.wait(contacts.map((val) async {
        AppResponse<Profile> userResponse = await read(val);
        if (userResponse.isSuccess) users.add(userResponse.data);
      }));
      return AppResponse.named(data: users.toList());
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<Profile>> getCurrent() async {
    try {
      String url = "/self";
      var response = await client.get(url);
      var json = jsonDecode(response.body);
      return AppResponse.named(
          data: Profile.fromJson(_convertJson(json["user"])));
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<Profile>> read(String userId) async {
    try {
      String url = "/users/$userId/";
      var response = await client.get(url);
      var json = jsonDecode(response.body);
      return AppResponse.named(
          data: Profile.fromJson(_convertJson(json["profile"])));
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  Map _convertJson(json) {
    var newJson = Map<String, dynamic>.from(json);
    newJson["name"] = json["fullname"];
    newJson["profilePic"] = json["avatar"];
    newJson["type"] = "user";
    newJson["id"] = json["user_id"];
    // Hack
    newJson["userId"] = json["phone"];
    return newJson;
  }

  @override
  Future<AppResponse<bool>> update(Profile user) async {
    try {
      String url = "/users/${user.id}/";
      await client.post(url, body: jsonEncode(user.toJson()));
      return AppResponse.named(data: true);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  Future<AppResponse<bool>> blockProfile(String userId, bool block) async {
    try {
      String url = "/users/$userId/${block ? "block" : "unblock"}";
      var response = await client.post(
        url,
      );
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint(
        "Error $e",
      );
      return AppResponse.named(error: e.toString());
    }
  }

  Future<AppResponse<List<Profile>>> getProfilesList(
      List<String> userIds) async {
    try {
      List<Profile> users = [];
      users = await Future.wait(userIds.map((e) async {
        AppResponse<Profile> userResponse = await read(e);
        if (userResponse.isSuccess) {
          return userResponse.data;
        }
      }).toList());
      users = users.where((element) => element != null).toList();
      return AppResponse.named(data: users);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<String>> create(Profile object) {
    // TODO: implement create
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> delete(String id) {
    // TODO: implement delete
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<Profile>> getByUser(String userId) {
    // TODO: implement getByUser
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<String>> getCurrentId() async {
    try {
      var response = await tokenRepository.getToken();
      if (response.isSuccess) {
        return AppResponse.named(data: response.data.userId);
      }
      return AppResponse.named(error: "Not found");
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<bool>> follow(String profileId, {bool follow}) {
    // TODO: implement follow
    throw UnimplementedError();
  }

  @override
  Stream<bool> onFollow() {
    // TODO: implement onFollow
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<ProfileFollowing>>> getFollowersFollowing(String id) {
    // TODO: implement getFollowersFollowing
    throw UnimplementedError();
  }

  @override
  Stream<ProfileFollowing> onFollowerChanged() {
    // TODO: implement onFollowerChanged
    throw UnimplementedError();
  }

  @override
  Stream<ProfileFollowing> onFollowingChanged() {
    // TODO: implement onFollowingChanged
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<Profile>>> getFollowers(String id) {
    // TODO: implement getFollowers
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<Profile>>> getFollowing(String id) {
    // TODO: implement getFollowing
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<Profile>>> getFriends(String id) {
    // TODO: implement getFriends
    throw UnimplementedError();
  }
}
